﻿using System;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Programmable {
  [Item("Single-objective ParameterVector Improver", "Improves a solution by calling GetNeighbors and Evaluate of the corresponding problem definition.")]
  [StorableClass]
  public sealed class SingleObjectiveParameterVectorImprover : SingleSuccessorOperator, ISingleObjectiveImprovementOperator, IStochasticOperator {
    public IValueLookupParameter<IItem> SolutionParameter {
      get { return new ValueLookupParameter<IItem>("notused"); }
    }

    public ILookupParameter<IRandom> RandomParameter {
      get { return (ILookupParameter<IRandom>)Parameters["Random"]; }
    }

    public ILookupParameter<ISingleObjectiveProblemDefinitionHost> ProblemDefinitionParameter {
      get { return (ILookupParameter<ISingleObjectiveProblemDefinitionHost>)Parameters["ProblemDefinition"]; }
    }

    public ILookupParameter<Configuration> ConfigurationParameter {
      get { return (ILookupParameter<Configuration>)Parameters["Configuration"]; }
    }

    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }

    public ILookupParameter<BoolValue> MaximizationParameter {
      get { return (ILookupParameter<BoolValue>)Parameters["Maximization"]; }
    }

    public IValueLookupParameter<IntValue> ImprovementAttemptsParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["ImprovementAttempts"]; }
    }

    public IValueLookupParameter<IntValue> SampleSizeParameter {
      get { return (IValueLookupParameter<IntValue>)Parameters["SampleSize"]; }
    }

    public ILookupParameter<IntValue> LocalEvaluatedSolutionsParameter {
      get { return (ILookupParameter<IntValue>)Parameters["LocalEvaluatedSolutions"]; }
    }

    [StorableConstructor]
    private SingleObjectiveParameterVectorImprover(bool deserializing) : base(deserializing) { }
    private SingleObjectiveParameterVectorImprover(SingleObjectiveParameterVectorImprover original, Cloner cloner) : base(original, cloner) { }
    public SingleObjectiveParameterVectorImprover() {
      Parameters.Add(new LookupParameter<IRandom>("Random", "The random number generator to use."));
      Parameters.Add(new LookupParameter<ISingleObjectiveProblemDefinitionHost>("ProblemDefinition", "The host that holds the problem definition."));
      Parameters.Add(new LookupParameter<Configuration>("Configuration", "An item that holds the problem's configuration."));
      Parameters.Add(new LookupParameter<DoubleValue>("Quality", "The quality of the parameter vector."));
      Parameters.Add(new LookupParameter<BoolValue>("Maximization", "Whether the problem should be minimized or maximized."));
      Parameters.Add(new ValueLookupParameter<IntValue>("ImprovementAttempts", "The number of improvement attempts the operator should perform.", new IntValue(100)));
      Parameters.Add(new ValueLookupParameter<IntValue>("SampleSize", "The number of samples to draw from the neighborhood function at maximum.", new IntValue(300)));
      Parameters.Add(new LookupParameter<IntValue>("LocalEvaluatedSolutions", "The number of solution evaluations that have been performed."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new SingleObjectiveParameterVectorImprover(this, cloner);
    }

    public override IOperation Apply() {
      var random = RandomParameter.ActualValue;
      var host = ProblemDefinitionParameter.ActualValue;
      if (host.Instance == null) throw new InvalidOperationException("Problem definition is not available.");
      var config = ConfigurationParameter.ActualValue;
      var maximize = MaximizationParameter.ActualValue.Value;
      var maxAttempts = ImprovementAttemptsParameter.ActualValue.Value;
      var sampleSize = SampleSizeParameter.ActualValue.Value;
      var vector = Helper.Extract(ExecutionContext.Scope, config);
      var quality = QualityParameter.ActualValue == null ? host.Instance.Evaluate(random, vector) : QualityParameter.ActualValue.Value;

      var count = 0;
      for (var i = 0; i < maxAttempts; i++) {
        ParameterVector best = null;
        var bestQuality = quality;
        foreach (var neighbor in host.Instance.GetNeighbors(random, vector).Take(sampleSize)) {
          var q = host.Instance.Evaluate(random, neighbor);
          count++;
          if (maximize && bestQuality > q || !maximize && bestQuality < q) continue;
          best = neighbor;
          bestQuality = q;
        }
        if (best == null) break;
        vector = best;
        quality = bestQuality;
      }

      LocalEvaluatedSolutionsParameter.ActualValue = new IntValue(count);
      Helper.Write(ExecutionContext.Scope, vector);
      return base.Apply();
    }
  }
}
